/* There's some trickly language in the GLX spec about how this is supposed
 * to work.  Basically, if a given component size is either not specified
 * or the requested size is zero, it is supposed to act like PERFER_SMALLER.
 * Well, that's really hard to do with the code as-is.  This behavior is
 * closer to correct, but still not technically right.
 */
#define PREFER_LARGER_OR_ZERO(comp)             \
  do {                                          \
    if ( ((*a)-> comp) != ((*b)-> comp) ) {     \
      if ( ((*a)-> comp) == 0 ) {               \
        return -1;                              \
      }                                         \
      else if ( ((*b)-> comp) == 0 ) {          \
        return 1;                               \
      }                                         \
      else {                                    \
        return ((*b)-> comp) - ((*a)-> comp) ;  \
      }                                         \
    }                                           \
  } while( 0 )

#define PREFER_LARGER(comp)                     \
  do {                                          \
    if ( ((*a)-> comp) != ((*b)-> comp) ) {     \
      return ((*b)-> comp) - ((*a)-> comp) ;    \
    }                                           \
  } while( 0 )

#define PREFER_SMALLER(comp)                    \
  do {                                          \
    if ( ((*a)-> comp) != ((*b)-> comp) ) {     \
      return ((*a)-> comp) - ((*b)-> comp) ;    \
    }                                           \
  } while( 0 )

  #define MATCH_DONT_CARE( param )        \
  do {                                  \
    if ( ((int) a-> param != (int) GLX_DONT_CARE)   \
      && (a-> param != b-> param) ) {        \
      return False;                             \
      }                                           \
      } while ( 0 )
      
      #define MATCH_MINIMUM( param )                  \
      do {                                          \
        if ( ((int) a-> param != (int) GLX_DONT_CARE)       \
          && (a-> param > b-> param) ) {         \
          return False;                             \
          }                                           \
          } while ( 0 )
          
          #define MATCH_EXACT( param )                    \
          do {                                          \
            if ( a-> param != b-> param) {              \
              return False;                             \
              }                                           \
              } while ( 0 )
              
              /* Test that all bits from a are contained in b */
              #define MATCH_MASK(param)                       \
              do {                                          \
                if ( ((int) a-> param != (int) GLX_DONT_CARE)       \
                  && ((a->param & ~b->param) != 0) ) {   \
                  return False;                             \
                  }                                           \
                  } while (0);
  
static void
init_fbconfig_for_chooser(struct glx_config * config,
                            GLboolean fbconfig_style_tags)
  {
    memset(config, 0, sizeof(struct glx_config));
    config->visualID = (XID) GLX_DONT_CARE;
    config->visualType = GLX_DONT_CARE;
    
    /* glXChooseFBConfig specifies different defaults for these properties than
     * glXChooseVisual.
     */
    if (fbconfig_style_tags) {
      config->rgbMode = GL_TRUE;
      config->doubleBufferMode = GLX_DONT_CARE;
      config->renderType = GLX_RGBA_BIT;
    }
    
    config->drawableType = GLX_WINDOW_BIT;
    config->visualRating = GLX_DONT_CARE;
    config->transparentPixel = GLX_NONE;
    config->transparentRed = GLX_DONT_CARE;
    config->transparentGreen = GLX_DONT_CARE;
    config->transparentBlue = GLX_DONT_CARE;
    config->transparentAlpha = GLX_DONT_CARE;
    config->transparentIndex = GLX_DONT_CARE;
    
    config->xRenderable = GLX_DONT_CARE;
    config->fbconfigID = (GLXFBConfigID) (GLX_DONT_CARE);
    
    config->swapMethod = GLX_DONT_CARE;
  }
  
  /**
   * Determine if two GLXFBConfigs are compatible.
   *
   * \param a  Application specified config to test.
   * \param b  Server specified config to test against \c a.
   */
  static Bool
  fbconfigs_compatible(const struct glx_config * const a,
                       const struct glx_config * const b)
  {
    MATCH_DONT_CARE(doubleBufferMode);
    MATCH_DONT_CARE(visualType);
    MATCH_DONT_CARE(visualRating);
    MATCH_DONT_CARE(xRenderable);
    MATCH_DONT_CARE(fbconfigID);
    MATCH_DONT_CARE(swapMethod);
    
    MATCH_MINIMUM(rgbBits);
    MATCH_MINIMUM(numAuxBuffers);
    MATCH_MINIMUM(redBits);
    MATCH_MINIMUM(greenBits);
    MATCH_MINIMUM(blueBits);
    MATCH_MINIMUM(alphaBits);
    MATCH_MINIMUM(depthBits);
    MATCH_MINIMUM(stencilBits);
    MATCH_MINIMUM(accumRedBits);
    MATCH_MINIMUM(accumGreenBits);
    MATCH_MINIMUM(accumBlueBits);
    MATCH_MINIMUM(accumAlphaBits);
    MATCH_MINIMUM(sampleBuffers);
    MATCH_MINIMUM(maxPbufferWidth);
    MATCH_MINIMUM(maxPbufferHeight);
    MATCH_MINIMUM(maxPbufferPixels);
    MATCH_MINIMUM(samples);
    
    MATCH_DONT_CARE(stereoMode);
    MATCH_EXACT(level);
    
    MATCH_MASK(drawableType);
    MATCH_MASK(renderType);
    
    /* There is a bug in a few of the XFree86 DDX drivers.  They contain
     * visuals with a "transparent type" of 0 when they really mean GLX_NONE.
     * Technically speaking, it is a bug in the DDX driver, but there is
     * enough of an installed base to work around the problem here.  In any
     * case, 0 is not a valid value of the transparent type, so we'll treat 0
     * from the app as GLX_DONT_CARE. We'll consider GLX_NONE from the app and
     * 0 from the server to be a match to maintain backward compatibility with
     * the (broken) drivers.
     */
    
    if (a->transparentPixel != (int) GLX_DONT_CARE && a->transparentPixel != 0) {
      if (a->transparentPixel == GLX_NONE) {
        if (b->transparentPixel != GLX_NONE && b->transparentPixel != 0)
          return False;
      }
      else {
        MATCH_EXACT(transparentPixel);
      }
      
      switch (a->transparentPixel) {
        case GLX_TRANSPARENT_RGB:
          MATCH_DONT_CARE(transparentRed);
          MATCH_DONT_CARE(transparentGreen);
          MATCH_DONT_CARE(transparentBlue);
          MATCH_DONT_CARE(transparentAlpha);
          break;
          
        case GLX_TRANSPARENT_INDEX:
          MATCH_DONT_CARE(transparentIndex);
          break;
          
        default:
          break;
      }
    }
    
    return True;
  }
  
  /**
   * Compare two GLXFBConfigs.  This function is intended to be used as the
   * compare function passed in to qsort.
   *
   * \returns If \c a is a "better" config, according to the specification of
   *          SGIX_fbconfig, a number less than zero is returned.  If \c b is
   *          better, then a number greater than zero is return.  If both are
   *          equal, zero is returned.
   * \sa qsort, glXChooseVisual, glXChooseFBConfig, glXChooseFBConfigSGIX
   */
  static int
  fbconfig_compare(struct glx_config **a, struct glx_config **b)
  {
    /* The order of these comparisons must NOT change.  It is defined by
     * the GLX 1.4 specification.
     */
    
    PREFER_SMALLER(visualSelectGroup);
    
    /* The sort order for the visualRating is GLX_NONE, GLX_SLOW, and
     * GLX_NON_CONFORMANT_CONFIG.  It just so happens that this is the
     * numerical sort order of the enums (0x8000, 0x8001, and 0x800D).
     */
    PREFER_SMALLER(visualRating);
    
    /* This isn't quite right.  It is supposed to compare the sum of the
     * components the user specifically set minimums for.
     */
    PREFER_LARGER_OR_ZERO(redBits);
    PREFER_LARGER_OR_ZERO(greenBits);
    PREFER_LARGER_OR_ZERO(blueBits);
    PREFER_LARGER_OR_ZERO(alphaBits);
    
    PREFER_SMALLER(rgbBits);
    
    if (((*a)->doubleBufferMode != (*b)->doubleBufferMode)) {
      /* Prefer single-buffer.
       */
      return (!(*a)->doubleBufferMode) ? -1 : 1;
    }
    
    PREFER_SMALLER(numAuxBuffers);
    
    PREFER_SMALLER(sampleBuffers);
    PREFER_SMALLER(samples);
    
    PREFER_LARGER_OR_ZERO(depthBits);
    PREFER_SMALLER(stencilBits);
    
    /* This isn't quite right.  It is supposed to compare the sum of the
     * components the user specifically set minimums for.
     */
    PREFER_LARGER_OR_ZERO(accumRedBits);
    PREFER_LARGER_OR_ZERO(accumGreenBits);
    PREFER_LARGER_OR_ZERO(accumBlueBits);
    PREFER_LARGER_OR_ZERO(accumAlphaBits);
    
    PREFER_SMALLER(visualType);
    
    /* None of the pbuffer or fbconfig specs say that this comparison needs
     * to happen at all, but it seems like it should.
     */
    PREFER_LARGER(maxPbufferWidth);
    PREFER_LARGER(maxPbufferHeight);
    PREFER_LARGER(maxPbufferPixels);
    
    return 0;
  }
  
  
static int
choose_visual(struct glx_config ** configs, int num_configs,
              const int *attribList, GLboolean fbconfig_style_tags)
{
   struct glx_config test_config;
   int base;
   int i;

   /* This is a fairly direct implementation of the selection method
    * described by GLX_SGIX_fbconfig.  Start by culling out all the
    * configs that are not compatible with the selected parameter
    * list.
    */

   init_fbconfig_for_chooser(&test_config, fbconfig_style_tags);
   __glXInitializeVisualConfigFromTags(&test_config, 512,
                                       (const INT32 *) attribList,
                                       GL_TRUE, fbconfig_style_tags);

   base = 0;
   for (i = 0; i < num_configs; i++) {
      if (fbconfigs_compatible(&test_config, configs[i])) {
         configs[base] = configs[i];
         base++;
      }
   }

   if (base == 0) {
      return 0;
   }

   if (base < num_configs) {
      (void) memset(&configs[base], 0, sizeof(void *) * (num_configs - base));
   }

   /* After the incompatible configs are removed, the resulting
    * list is sorted according to the rules set out in the various
    * specifications.
    */

   qsort(configs, base, sizeof(struct glx_config *),
         (int (*)(const void *, const void *)) fbconfig_compare);
   return base;
}

